/* root.cpp -- Bairstow's complex root finder
 *
 *  (C) 1991, 1998, C. Bond
 *
 *  Finds all real and complex roots of polynomials
 *  with real coefficients.
 *
 *  Polynomial coefficient order is from high to low:
 *
 *      a[0]      a[1]      ...       a[n]  
 *      p^n     c1*p^(n-1)             cn
 *
 *  Features:
 *      o Global, self-starting method,
 *      o Does not require initial estimates,
 *      o Finds all roots and quadratic factors,
 *
 *  Caveats:
 *      All polynomial root finders have problems
 *      maintaining accuracy in the presence of
 *      repeated roots.
 *      This is no exception!
 */
#include <stdio.h>
#include <math.h>
#include <complex.h>

#define MAX_TERMS 21
extern int precision_error_flag;
double b[MAX_TERMS],c[MAX_TERMS];

void find_quad_factor(double a[],int n)
{
		double r,s,dn,dr,ds,drn,dsn,eps;
	int i,iter;

	r = s = 0;
	dr = 1.0;
	ds = 0;
	eps = 1e-14;
	iter = 1;

	while ((fabs(dr) + fabs(ds)) > eps) {
		if ((iter % 200) == 0) {
			r = (double)rand()/16000.;
		}
		if ((iter % 500) == 0) {
			eps *= 10.0;
			precision_error_flag = 1;
			printf("Loss of precision\n");
		}
		b[1] = a[1] - r;
		c[1] = b[1] - r;

		for (i=2;i<=n;i++){
			b[i]=a[i] - r * b[i-1] -s * b[i-2];
			c[i]=b[i] - r * c[i-1] -s * c[i-2];
		}
		dn = c[n-1] * c[n-3] - c[n-2] * c[n-2];
		drn = b[n] * c[n-3] - b[n-1] * c[n-2];
		dsn = b[n-1] * c[n-1] - b[n] * c[n-2];

		if (fabs(dn) < 1e-16) {
			dn = 1;
			drn = 1;
			dsn = 1;
		}
		dr = drn / dn;
		ds = dsn / dn;

		r += dr;
		s += ds;
		iter++;
	}
	for (i=0;i<n-1;i++) a[i] = b[i];
	a[n] = s;
	a[n-1] = r;
}

void Roots(double a[],int n,complex<double> p[])
{
    complex<double> bb,cc;
    complex<double> disc;
	double tmp;
	int i;

	b[0] = c[0] = 1.0;

/* ensure that leading coefficient is unity */
	if ((tmp = a[0]) != 1.0) {
		for (i=0;i<=n;i++) {
			a[i] /= tmp;
		}
	}
/* handle all zero roots */
	while ((n > 2) && (fabs(a[n]) < 1e-6)) {  
		p[n-1] = 0.0;
		n--;
	}
/* find quadratic factors */
	while (n > 2) {
		find_quad_factor(a,n);
        bb = complex<double>(a[n-1]);
        cc = complex<double>(a[n]);
        disc = sqrt(bb*bb-4.0*cc)/2.0;        
		p[n-2] = -bb/2.0-disc;
		p[n-1] = -bb/2.0+disc;
		n -= 2;
	}
/* last roots may be a complex pair or single, real root */
	if (n == 2) {
        bb = complex<double>(a[1]);
        cc = complex<double>(a[2]);
        disc = sqrt(bb*bb-4.0*cc)/2.0;
		p[0] = -bb/2.0-disc;
		p[1] = -bb/2.0+disc;
	}
	else p[0] = -a[1];
}

